
import os
import random

from PyTASTileType import *
from PyTASAssembly import *


NBRS = [(0,1,0),(1,0,0),(0,-1,0),(-1,0,0),(0,0,1),(0,0,-1)]
#==============================
# N = 0
# E = 1
# S = 2
# W = 3
# U = 4
# D = 5
#=============================#
OPP_DIR = {0:2, 1:3, 2:0, 3:1, 4:5, 5:4}

EMPTY_LOC           = 0
FRONTIER_LOC        = 1
FRONTIER_LOC_OFF    = 2


class PyTAS3DAssembly(PyTASAssembly):
    def __init__(self, tileWidths=100, tileHeights=100):
        PyTASAssembly.__init__(self, tileWidths, tileHeights)
        
        self.NBRS = [(0,1,0),(1,0,0),(0,-1,0),(-1,0,0),(0,0,1),(0,0,-1)]
        self.DIRSTRS = ["N", "E", "S", "W", "U", "D"]
        self.OPPDIRS = [2, 3, 0, 1, 5, 4]

        self.grid = dict()
        self.tileTypes = dict()        
        self.seedPatch = []
        self.frontier = set()
        self.temperature = 2
        self.numTiles = 0
        self.tileSystemDef = None
        self.FrontierHashTable = dict()

    def ClearSpace(self):
        newAssembly = PyTAS3DAssembly()
        del self.grid
        self.grid = newAssembly.grid
        self.frontier = set()
        self.numTiles = 0
        self.FrontierHashTable = dict()
        
    # Parameters growIfNeeded is unused
    def GetLoc(self, coords, growIfNeeded=False, includeFrontier=False):
        if coords in self.grid:
            loc = self.grid[coords]
        else:
            return None
        if includeFrontier == False and (loc == FRONTIER_LOC or loc == FRONTIER_LOC_OFF):
            return None
        else:
            return loc
    
    # Parameter growIfNeeded is unused
    def SetLoc(self, coords, ttype, growIfNeeded=False):
        if ttype == FRONTIER_LOC or ttype == FRONTIER_LOC_OFF:
            self.grid[coords] = ttype
        else:
            self.grid[coords] = Tile(ttype, coords)
        return True
        
    def ClearLoc(self, coords):
        del self.grid[coords]
        return True
        
    #===========================================================================
    # def AddTile(self, ttype, coords):
    #     loc = self.GetLoc(coords)
    #     if not loc is None:
    #         SimUtils.globalApp.OutputMessage("ERROR: Trying to add another node at location " + str(coords))
    #         return False
    #     if self.SetLoc(coords, ttype) == False:
    #         return False
    #     emptyNbrLocs = self.GetNbrLocs(coords, True)
    #     #print("unoccupied nbrs of " + str(coords) + ": " + str(emptyNbrLocs))
    #     for n in emptyNbrLocs:   # loop over unoccupied nbr locations to find occupied nbrs of each and build a list of fitting types
    #         self.AddToFrontierIfNecessary(n)
    #     self.numTiles = self.numTiles + 1
    #     return True
    #===========================================================================

    #===========================================================================
    # def AddToFrontierIfNecessary(self, coords):
    #     fittingNbrSet = self.GetFittingTilesForLoc(coords)
    #     if fittingNbrSet is None:
    #         return
    #     fLoc = FrontierLocation(coords, fittingNbrSet)
    #     self.frontier.discard(fLoc)
    #     self.frontier.add(fLoc)
    #     #print("fitting nbr set: " + str(fittingNbrSet))
    #===========================================================================


    def RemoveTile(self, coords):
        if not self.GetLoc(coords) is None:
            self.ClearLoc(coords)
            self.numTiles = self.numTiles - 1

            # Check to see if this location should now be added to the frontier
            self.AddToFrontierIfNecessary(coords)

            # Check to see if nbrs were in the frontier but should now be removed
            nbrLocs = self.GetNbrLocs(coords, True)
            for n in nbrLocs:
                fLoc = FrontierLocation(n, None)
                self.frontier.discard(fLoc)
                self.AddToFrontierIfNecessary(n)
            return True
        else:
            print("ERROR: Trying to remove nonexistent node at location" + str(coords))
            return False


    def GetNbrLocs(self, coords, bEmpty=False):
        retlist = []
        for n in NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1], coords[2]+n[2])
            if bEmpty == False:
                retlist.append(nbrCoords)
            elif self.GetLoc(nbrCoords) is None:
                retlist.append(nbrCoords)
        return retlist


    def GetOccupiedNbrLocs(self, coords):
        retlist = []
        for n in NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1], coords[2]+n[2])
            nbrNode = self.GetLoc(nbrCoords)
            if not nbrNode is None:
                retlist.append(nbrNode)
        return retlist
    
#===============================================================================
#     def GetFittingTilesForLoc(self, coords):
#         retSet = None
#         count = 0
#         nbrGlues = []
#         for n in NBRS:
#             nbrCoords = (coords[0]+n[0], coords[1]+n[1], coords[2]+n[2])
#             nbrTile = self.GetLoc(nbrCoords)
#             if not nbrTile is None:
#                 oppDir = OPP_DIR[count]
#                 nbrGlues.append(nbrTile.type.sides[oppDir].Hash())
#                 #if retSet is None:
#                 #    retSet = nbrTile.type.matches[oppDir]
#                 #else:
#                 #    retSet = retSet.intersection(nbrTile.type.matches[oppDir])
#             else:
#                 nbrGlues.append(None)
#             count = count + 1
# 
#         signature = str(nbrGlues)
#         if signature in self.FrontierHashTable:
#             return self.FrontierHashTable[signature]
# 
#         for t in self.tileTypes:
#             totStr = 0
#             for n in range(len(NBRS)):
#                 if not nbrGlues[n] is None:
#                     if nbrGlues[n] == self.tileTypes[t].sides[n].Hash():
#                         totStr = totStr + self.tileTypes[t].sides[n].strength
#             if totStr >= self.temperature:
#                 if retSet is None:
#                     retSet = set()
#                 retSet.add(t)
#         self.FrontierHashTable[signature] = retSet
#         return retSet
#===============================================================================

    def StepForward(self, full=False):
        # Pick a spot from the frontier, add a tile
        if len(self.frontier) == 0:
            SimUtils.globalApp.OutputMessage("The frontier is empty!")
            return None
        elif not full:
            frontierLoc = random.sample(self.frontier, 1)[0]
            self.frontier.remove(frontierLoc)
            typeName = random.sample(frontierLoc.fittingTypeNames, 1)[0]
            
            if self.AddTile(self.tileTypes[typeName], frontierLoc.coords)[0] == True:
                return (frontierLoc.coords, typeName)

            SimUtils.globalApp.OutputMessage("Error adding tile to location" + str(frontierLoc.coords))
            return None
        else:
            added = []
            frontierList = list(self.frontier)
            for frontierLoc in frontierList:
                self.frontier.remove(frontierLoc)
                typeName = random.sample(frontierLoc.fittingTypeNames, 1)[0]

                if self.AddTile(self.tileTypes[typeName], frontierLoc.coords)[0] == True:
                    added.append((frontierLoc.coords, typeName))
                else:
                    SimUtils.globalApp.OutputMessage("Error adding tile to location" + str(frontierLoc.coords))
            if not added:
                return None
            else:
                return added
                
    def PlaceSeed(self):
        if self.bHas3D == False:
            for ttName in self.tileTypes:
                self.tileTypes[ttName].SetDimensions(3)
            self.bHas3D = True
        SimUtils.globalApp.OutputMessage("Placing seed patch: " + str(self.seedPatch))
        for seedTile in self.seedPatch:
            self.AddTile(self.tileTypes[seedTile[0]], seedTile[1])
        

    def AssemblyToString(self, tileList=None):
        assemblyString = self.tileSystemDef['tdsFile'] + '\n'
        assemblyString += 'Temperature=' + str(self.tileSystemDef['temperature']) + '\n'

        for tile in self.grid.itervalues():
            assemblyString += tile.type.name + ' '
            assemblyString += str(tile.coords[0]) + ' ' + str(tile.coords[1]) + ' ' + str(tile.coords[2]) + '\n'

        return assemblyString

        
    def CreateTileTypeImages(self, backgroundColor=None):
        # TODO: load cubemap textures
        pass